#ifndef _PDP_COMMON_H_INCLUDED_
#define _PDP_COMMON_H_INCLUDED_

/**
 * @file
 * общие функции для работы с метками безопасности
 */

#ifdef __cplusplus
extern "C" {
#endif

#ifdef __KERNEL__
	#include <linux/types.h>

	#define PDP_USRKRNL(u,k) k
	#define PDP_KRNL(x) x
	#define PDP_NKRNL(x)
	#define PDP_KRNL_ARG(x) , x
#else
	#include <inttypes.h> /* 4 uint*_t */
	#include <stdlib.h> /* 4 size_t */

	#define PDP_USRKRNL(u,k) u
	#define PDP_KRNL(x)
	#define PDP_NKRNL(x) x
	#define PDP_KRNL_ARG(x)
#endif

typedef uint8_t		PDP_LEV_T;
#define PDP_PRNTF_LEV    PRIu8
#define PDP_SCNF_LEV     SCNu8

typedef uint8_t		PDP_ILEV_T;
#define PDP_PRNTF_ILEV    SCNu8
#define PDP_SCNF_ILEV     SCNu8

typedef uint64_t	PDP_CAT_T;
#define PDP_PRNTF_CAT    PRIx64
#define PDP_SCNF_CAT     SCNx64

typedef uint16_t	PDP_TYPE_T;
#define PDP_PRNTF_TYPE    PRIx16
#define PDP_SCNF_TYPE     SCNx16

typedef uint32_t	PDP_CNT_SIZE_T;
#define PDP_PRNTF_CNT_SIZE    PRIx32
#define PDP_SCNF_CNT_SIZE     SCNx32

typedef uint64_t	PDP_RSEED_T;
#define PDP_PRNTF_RSEED     PRIx64
#define PDP_SCNF_RSEED     SCNu64

#define PDPT_CCNR		0x01
#define PDPT_CCNRI		0x02
#define PDPT_EHOLE		0x04

//#define PDPT_CONTAINER	0x08
//#define PDPT_SUBJECT		0x10
//#define PDPT_OBJECT			0x20

typedef struct PDP_ROLE_T /* struct role for subjects and objects*/
{
	uint64_t	seed;	/* inode number */
	uint32_t	uid;	/* uid_t (for subjects) */
	uint8_t		mode;	/* mode mask (for objects) */
} PDP_ROLE_T; /*size 16*/

/* flags */
//#define PDPF_ROLE		0x01
//#define PDPF_AROLE		0x02

#ifdef __KERNEL__
	#define PDPL_PRINTFA(l) PDPML_PRINTFA(l->ml)
//	#define PDPL_PRINTFA(l) pdpl_lev(l),pdpl_ilev(l),pdpl_cat(l),pdpl_type(l)

	#define PDPL_PRINTF  "%" "hhu"  ":%"  "hhu" ":%" "llx" ":%" "hx"
	#define PDPML_PRINTFA(ml) ml->lev,ml->ilev,ml->cat,ml->type
#else
	#define PDPL_PRINTFA(l) pdpl_lev(l),pdpl_ilev(l),pdpl_cat(l),pdpl_type(l)
	#define PDPL_PRINTF "%" PDP_PRNTF_LEV  ":%" PDP_PRNTF_ILEV ":%" PDP_PRNTF_CAT ":%" PDP_PRNTF_TYPE

	#define PDPML_PRINTFA(l) pdpml_lev(l),pdpml_ilev(l),pdpml_cat(l),pdpml_type(l)
#endif

	#define PDPML_PRINTF PDPL_PRINTF

/**
 * @brief Определение метки безопасности в соттветсвии с ДП моделью для объектов и субъектов доступа (PDP - Parsec Dostup Potoky)
 * Работа с экземплярами данного типа ведется только через указатель, внутреннее состояние доступно только через
 * вызов определенной функции (инкапсуляция)
 */
typedef struct PDP_LABEL_T			PDPL_T;

/**
 * @brief Определение мандатной метки в соттветсвии с ДП моделью для объектов и субъектов доступа (PDP - Parsec Dostup Potoky)
 * Работа с экземплярами данного типа ведется только через указатель, внутреннее состояние доступно только через
 * вызов определенной функции (инкапсуляция)
 */
typedef struct PDP_MAC_LABEL_T		PDPML_T;

/**
 * @brief Определение ролевой метки безопасности в соттветсвии с ДП моделью для объектов и субъектов доступа (PDP - Parsec Dostup Potoky)
 * Работа с экземплярами данного типа ведется только через указатель, внутреннее состояние доступно только через
 * вызов определенной функции (инкапсуляция)
 */
typedef struct PDP_ROLES_LABEL_T	PDPRL_T;

#define PDPML_ZERO  (pdpl_ml(PDPL_ZERO))
#define PDPRL_EMPTY (pdpl_rl(PDPL_ZERO))

//extern const PDPML_T *PDPML_ZERO; /**< Нулевая метка безопасности: уровень конф - 0, целостности - низкий; категорий нет; флагов нет */
//extern const PDPML_T *PDPML_EQU; /**< Метка безопасности объекта дырки: уровень конф - 0, целостности - низкий; категорий нет; флагов нет */

extern const PDPL_T *PDPL_ZERO; /**< Нулевая метка безопасности: уровень конф - 0, целостности - низкий; категорий нет; флагов нет; ролей нет */
extern const PDPL_T *PDPL_EQU; /**< Метка безопасности объекта дырки: уровень конф - 0, целостности - низкий; категорий нет; флагов нет; ролей нет */

const PDPL_T* pdpl_get(const PDPL_T *l);
PDPL_T* pdplnc_get(PDPL_T *l);
const PDPML_T* pdpml_get(const PDPML_T *ml);
const PDPRL_T* pdprl_get(const PDPRL_T *rl);

/**
 * Выделение памяти и инициализация метки безопасности
 * @return указатель на метку, которая представляет собой копию нулевой метки (),
 * в случае ошибки выделения памяти NULL
 */
PDPL_T* pdpl_get_new(PDP_USRKRNL(void,gfp_t f));

PDPML_T* pdpml_get_new(PDP_USRKRNL(void,gfp_t f));

PDPRL_T* pdprl_get_new(PDP_USRKRNL(void,gfp_t f));

/* pdpXl_get_new_init */

PDPL_T* pdpl_get_new_init_mac(PDP_LEV_T lev, PDP_ILEV_T ilev, PDP_CAT_T cat, PDP_TYPE_T type PDP_KRNL_ARG(gfp_t f));

/**
 * @brief Выделение памяти и инициализация мандатной метки безопасности
 *
 * @param lev
 * @param ilev
 * @param cat
 * @param type
 * @return указатель на метку, в случае ошибки выделения памяти NULL
 */
PDPML_T* pdpml_get_new_init(PDP_LEV_T lev, PDP_ILEV_T ilev, PDP_CAT_T cat, PDP_TYPE_T type PDP_KRNL_ARG(gfp_t f));

PDPL_T* pdpl_get_new_init_mac_roles(PDP_LEV_T lev, PDP_ILEV_T ilev, PDP_CAT_T cat, PDP_TYPE_T type,
									const PDP_ROLE_T *rls, PDP_CNT_SIZE_T rlscnt, const PDP_ROLE_T *arls, PDP_CNT_SIZE_T arlscnt PDP_KRNL_ARG(gfp_t f));
PDPL_T* pdpl_get_new_init_mr(const PDPML_T *ml, const PDPRL_T *rl PDP_KRNL_ARG(gfp_t f));

PDPRL_T* pdprl_get_new_init_roles(const PDP_ROLE_T *rls, PDP_CNT_SIZE_T rlscnt, const PDP_ROLE_T *arls, PDP_CNT_SIZE_T arlscnt PDP_KRNL_ARG(gfp_t f));
/**
 * @brief Выделение памяти и инициализация списка ролей
 *
 * @return указатель на метку, в случае ошибки выделения памяти NULL
 */
PDPRL_T* pdprl_get_new_init( void* somearg PDP_KRNL_ARG(gfp_t f));

/**
 * @brief Выделение памяти и инициализация структуры роли или административной роли
 *
 * @param seed уникальный идентификатор роли
 * @param uid идентиикатор пользователя для инд. роли или 0 в противном случае
 * @param mode права доступа предоставляемые посредством данной роли (для объектов и ролей)
 * @return указатель на структуру, в случае ошибки выделения памяти NULL
 */
PDP_ROLE_T* pdpr_get_new_init(uint64_t seed, uint32_t uid, uint8_t mode PDP_KRNL_ARG(gfp_t f));

void pdp_roles_normalize(PDP_ROLE_T *roles, PDP_CNT_SIZE_T *roles_cnt);

/**
Освобождение ресурсов

@param l указатель на метку безопасности
*/
void pdpl_put(const PDPL_T* l);
void pdpml_put(const PDPML_T* ml);
void pdprl_put(const PDPRL_T* rl);

/**
 * @brief получение количества ссылок на метку безопасности, только для отладки
 * @param l указатель на исходную метку безопасности
 * @return количество ссылок на метку безопасности
*/
size_t pdpl_ucnt(const PDPL_T* l);
size_t pdpml_ucnt(const PDPML_T* ml);
size_t pdprl_ucnt(const PDPRL_T* rl);

///const PDPL_T* pdpl_get(PDP_O* l);

/**
 * @brief Дублирование метки безопасности
 * Создание ПОЛНОЙ копии исходной метки безопасности в динамически выделенной области памяти
 * @param l указатель на исходную метку безопасности
 * @return указатель на копию исходной метки безопасности, в случае ошибки выделения памяти NULL
*/
PDPL_T* pdpl_dup(const PDPL_T* l  PDP_KRNL_ARG(gfp_t f));
PDPML_T* pdpml_dup(const PDPML_T* l PDP_KRNL_ARG(gfp_t f));

/**
 * Создание копии исходного множества ролей в динамически выделенной области памяти
 * @param r указатель на исходное множество ролей
 * @param r_cnt количество ролей в исходном множестве
 * @return указатель на копию исходного множества ролей, в случае ошибки выделения памяти NULL
*/
struct PDP_ROLE_T* pdp_role_dup(const struct PDP_ROLE_T* r, uint32_t r_cnt PDP_KRNL_ARG(gfp_t f));

/**
 * Создание копии исходной метки rback в динамически выделенной области памяти
 * @param rl указатель на исходною метку rback
 * @return указатель на копию исходной метки rback, в случае ошибки выделения памяти NULL
*/
PDPRL_T* pdprl_dup(const PDPRL_T* rl  PDP_KRNL_ARG(gfp_t f));

/**
 * Добавление(в порядке убывания) роли new_role в массив ролей dest_role_mass (до этого ее несодержащий), содержащий role_count ролей
 * @param new_role указатель роль, которую необходимо добавить
 * @param dest_role_mass указатель на мастив ролей, в который необходимо добавить новую роль
 * @param role_count - число ролей в целевом массиве *
 * @return указатель на массив ролей, полученный после добаления новой роли, значение dst_role_count устанавливается в количество ролей в вновь созданном массиве,
 * в случае ошибки выделения памяти  NULL
 */
extern PDP_ROLE_T* rback_only_new_role_add(struct PDP_ROLE_T* new_role, struct PDP_ROLE_T* dest_role_mass, PDP_CNT_SIZE_T* role_count PDP_KRNL_ARG(gfp_t f));

/**
 * @brief обновление метки безопасности
 * Создание ПОЛНОЙ копии исходной метки безопасности @src в @dst в соответсвии с концепцией RCU(D):
 *  - уменьшение счетчика ссылок метки *dsc (pdpl_put(*dsc);)
 *  - cоздание полной копии исходной метки *src (*dsc=pdpl_dup(src);)
 *
 * Результирующая метка гарантировано имеет ссчетчик ссылок равный 1;
 *
 * Данная операция потокоНЕбезопасная, требуется захват блокировки, сответвующей указателю @dst.
 *
 * @param src указатель на исходную метку безопасности
 * @param dst указатель на указатель на результирующую метку безопасности
 * @return в случае успеха *dsc, т.е. указатель на копию исходной метки безопасности;
 * в случае ошибки выделения памяти NULL: если src==dst errno выставляется в EINAVL, в случае ошибки выделения памяти - ENOMEM
*/
PDPL_T* pdpl_upd(PDPL_T **dst,const PDPL_T *src PDP_KRNL_ARG(gfp_t f) );

int pdpr_cpy(PDP_ROLE_T *dst,const PDP_ROLE_T *src, uint32_t src_cnt PDP_KRNL_ARG(gfp_t f));
int pdprl_cpy(PDPRL_T *dst,const PDPRL_T *src PDP_KRNL_ARG(gfp_t f));


const PDPML_T* pdpl_ml(const PDPL_T* l);
const PDPRL_T* pdpl_rl(const PDPL_T* l);

PDP_LEV_T  pdpl_lev(const PDPL_T* l);
PDP_ILEV_T pdpl_ilev(const PDPL_T* l);
PDP_CAT_T  pdpl_cat(const PDPL_T* l);
PDP_TYPE_T pdpl_type(const PDPL_T* l);

void pdpl_set_lev(PDPL_T* l, PDP_LEV_T lev);
void pdpl_set_ilev(PDPL_T* l, PDP_ILEV_T ilev);

void pdpl_set_cat(PDPL_T* l, PDP_CAT_T cat);
void pdpl_and_cat(PDPL_T* l, PDP_CAT_T cat);
void pdpl_or_cat(PDPL_T* l, PDP_CAT_T cat);

void pdpl_set_type(PDPL_T* l, PDP_TYPE_T type);
void pdpl_and_type(PDPL_T* l, PDP_TYPE_T type);
void pdpl_or_type(PDPL_T* l, PDP_TYPE_T type);

PDP_LEV_T  pdpml_lev(const PDPML_T* l);
PDP_ILEV_T pdpml_ilev(const PDPML_T* l);
PDP_CAT_T  pdpml_cat(const PDPML_T* l);
PDP_TYPE_T pdpml_type(const PDPML_T* l);

void pdpml_set_lev(PDPML_T* l, PDP_LEV_T lev);
void pdpml_set_ilev(PDPML_T* l, PDP_ILEV_T ilev);

void pdpml_set_cat(PDPML_T* l, PDP_CAT_T cat);
void pdpml_and_cat(PDPML_T* l, PDP_CAT_T cat);
void pdpml_or_cat(PDPML_T* l, PDP_CAT_T cat);

void pdpml_set_type(PDPML_T* l, PDP_TYPE_T type);
void pdpml_and_type(PDPML_T* l, PDP_TYPE_T type);
void pdpml_or_type(PDPML_T* l, PDP_TYPE_T type);

/**
Сравнение меток безопасности

@param a указатель на метку безопасности
@param b указатель на метку безопасности
@param flags указатель на метку безопасности

@return 1 - если равны, 0 - если не равны
*/
int pdpl_cmp(const PDPL_T* a, const PDPL_T* b, int flags);
int pdpml_cmp(const PDPML_T* a, const PDPML_T* b, int flags);
int pdprl_cmp(const PDPRL_T* a, const PDPRL_T* b, int flags);

int pdp_role_cmp(const PDP_ROLE_T *a, const PDP_ROLE_T *b);

int pdpml_notg(const PDPML_T *less, const PDPML_T *great);
int pdpl_notg(const PDPL_T *less, const PDPL_T *great);

/**
Проверка, может ли объект-родитель с меткой @param parent иметь дочерний
объект с меткой @param child

@param child указатель на метку безопасности
@param parent указатель на метку безопасности

@return 0 если может, иначе не 0
*/
int pdp_child_parent(const PDPL_T *child, const PDPL_T *parent);

int pdpml_child_parent(const PDPML_T *child, const PDPML_T *parent);

/**
@breif Сравнение меток безопасности

Предосваляется для удобства, является надстройкой над функцией pdpl_cmp
static inline int pdpl_equ(const PDPL_T* a, const PDPL_T* b) { return 1==pdpl_cmp(a,b,0); }

@param a указатель на метку безопасности
@param b указатель на метку безопасности
@return не 0 - если равны, 0 - если не равны
*/
static inline int pdpl_equ(const PDPL_T* a, const PDPL_T* b) { return 1==pdpl_cmp(a,b,0); }
static inline int pdpml_equ(const PDPML_T* a, const PDPML_T* b) { return 1==pdpml_cmp(a,b,0); }
//TODOstatic inline int pdprl_equ(const PDPRL_T* a, const PDPRL_T* b) { return pdprl_cmp(a,b,0); }

static inline int pdpl_is_zero(const PDPL_T* l) { return 1==pdpl_equ(l,PDPL_ZERO); }
static inline int pdpml_is_zero(const PDPML_T* ml) { return 1==pdpml_equ(ml,PDPML_ZERO); }
//TODOstatic inline int pdprl_is_zero(const PDPRL_T* l) { return 1==pdpl_equ(l,PDPML_ZERO); }

/* int	pdp_isvalid(const PDP_O* l); */

/**
Получение метки безопасности из бинарного представления, метка будет выделена посредсвом pdp_alloc

@param buf указатель на буфер, который содержти метку безопасности в бинарном виде
@param size размер буфера
@param __out size_without_role размер буфера, относящийся к манд. атрибутам
@return указатель на метку безопасности или NULL в случае ошибки (неверный формат метки)
*/
PDPL_T* pdpl_get_from_raw(const void *buf, size_t size PDP_KRNL_ARG(gfp_t f));

PDPML_T* pdpml_get_from_raw(const void *raw, size_t size, size_t* size_without_role PDP_KRNL_ARG(gfp_t f));

/**
Получение метки rback из бинарного представления метки rback, память будет выделена посредсвом pdp_alloc

@param buf указатель на буфер, который содержти список ролей в бинарном виде
@param size размер буфера
@return указатель на метку ролей или NULL в случае ошибки (неверный формат метки)
*/
PDPRL_T* pdprl_get_from_raw_role(const void *buf, size_t size PDP_KRNL_ARG(gfp_t f));
/**
Получение метки rback из бинарного представления метки безопасности, память будет выделена посредсвом pdp_alloc

@param buf указатель на буфер, который содержти всю метку (включая манд.) в бинарном виде
@param size размер буфера
@param size_without_role размер буфера, относящийся к манд. атрибутам
@return указатель на метку ролей или NULL в случае ошибки (неверный формат метки)
*/
PDPRL_T* pdprl_get_from_raw_lbl(const void *buf, size_t size, size_t size_without_role PDP_KRNL_ARG(gfp_t f));

/**
Получение метки безопасности из бинарного представления

@param l указатель на метку безопасности
@param buf указатель на буфер, который содержти метку безопасности в бинарном виде
@param size размер буфера
@return 0 в случае успеха
*/
int pdpl_from_raw(PDPL_T* l, const void *buf, size_t size);

/**
Получение бинарного представления метки безопасности

@param l указатель на метку безопасности
@param size указатель на переменную, куда возвращается размер выделенного буфера (НЕ NULL)
@return указатель на буфер, выделенный функцией malloc(), который содержти метку безопасности в бинарном виде или NULL в случае ошибки
*/
void* pdpl_get_raw(const PDPL_T* l, size_t* size PDP_KRNL_ARG(gfp_t f)); /* буфер будет выделен функцией */
void* pdpml_get_raw(const PDPML_T* l, size_t* size PDP_KRNL_ARG(gfp_t f)); /* буфер будет выделен функцией */

/**
Получение бинарного представления метки rback

@param l указатель на метку rback
@param size указатель на переменную, куда возвращается размер выделенного буфера (НЕ NULL)
@return указатель на буфер, выделенный функцией malloc(), который содержти метку rback в бинарном виде или NULL в случае ошибки
*/
void* pdprl_get_raw(const PDPRL_T* ml, size_t* size PDP_KRNL_ARG(gfp_t f)); /* буфер будет выделен функцией */

extern const size_t PDPL_RAW_MIN_SIZE;

/**
Получение размера в байтах бинарного представления метки безопасности

@param l указатель на метку безопасности
@return размер в байтах бинарного представления метки безопасности
*/
size_t pdpl_raw_size(const PDPL_T* l);
size_t pdprawl_size(const void* rawl);

size_t pdpml_raw_size(const PDPML_T* ml);
size_t pdprawml_size(const void* rawml);

size_t pdprl_raw_size(const PDPRL_T* rl);
size_t pdprawrl_size(const void* rawrl);

size_t pdp_roles_raw_size(size_t cnt);
size_t pdp_roles_size(size_t cnt);

typedef struct _parsec_mac_raw_t parsec_mac_raw_t;
void pdpl_2_legacy_raw(const PDPL_T* l, parsec_mac_raw_t* mr);
PDPL_T* pdpl_get_from_legacy_raw(parsec_mac_raw_t* mr PDP_KRNL_ARG(gfp_t f));

//	extern size_t	pdpl_size(PDP_O * l); /* get external size */

#ifdef __KERNEL__
	#define R_OK MAY_READ
	#define W_OK MAY_WRITE
	#define X_OK MAY_EXEC
	#define A_OK MAY_APPEND
	#define C_OK MAY_CREATE
#else
	#define A_OK (R_OK<<1) /* APPEND need description */
	#define C_OK (A_OK<<1)
#endif
	#define PDP_O_CONTAINER   0x08000
	#define LEGACY_IGNORE_LEV 0x10000
	#define LEGACY_IGNORE_CAT 0x20000
	#define PDP_IGN_INT       0x40000

/**
Проверка доступа объека с субъекту

@param o указатель на метку безопасности объекта
@param s указатель на метку безопасности субъекта
@param mode режим доступа X_OK, R_OK, W_OK

@return 0 - доступ разрешен, иначе - запрещен
*/
int pdpl_permission(const PDPL_T* s, const PDPL_T* o, int mode); /* mode: X_OK, R_OK, W_OK or MAY_READ, MAY_WRITE, MAY_EXEC 4 kernel */

int pdpml_permission(const PDPML_T *s, const PDPML_T *o, int mode); /* mode: X_OK, R_OK, W_OK or MAY_READ, MAY_WRITE, MAY_EXEC 4 kernel */

int pdprl_permission(const PDPRL_T *s, const PDPRL_T *o, int mode);

#ifdef __cplusplus
}
#endif

#endif /* _PDP_COMMON_H_INCLUDED_ */
